/*
 * MeshBasedNode.cpp
 *
 * Copyright (C) 1999 Stephen F. White, 2004 J. "MUFTI" Scheurich
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
# include <fcntl.h>
#endif

#include "swt.h"
#include "stdafx.h"
#include "MeshBasedNode.h"
#include "Scene.h"
#include "FieldValue.h"
#include "SFInt32.h"
#include "Mesh.h"
#include "Face.h"
#include "MFFloat.h"
#include "MFInt32.h"
#include "MFVec2f.h"
#include "MFVec3f.h"
#include "SFNode.h"
#include "SFBool.h"
#include "SFVec3f.h"
#include "Vec2f.h"
#include "Vec3f.h"
#include "RenderState.h"
#include "DuneApp.h"
#include "NodeCoordinate.h"
#include "NodeNormal.h"
#include "NodeTextureCoordinate.h"
#include "NodeColor.h"
#include "NodeIndexedFaceSet.h"
#include "NodeTriangleSet.h"
#include "NodeElevationGrid.h"
#include "NodeGeoElevationGrid.h"

MeshBasedNode::MeshBasedNode(Scene *scene, Proto *proto)
  : GeometryNode(scene, proto)
{
    _isMesh = true;
    _mesh = NULL;
    _meshDirty = true;
}

MeshBasedNode::~MeshBasedNode()
{
    delete _mesh;
}


void
MeshBasedNode::meshDraw()
{
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }

    if (!_mesh) return;

    _mesh->sort();
    _mesh->draw();
}

Node * 
MeshBasedNode::toIndexedFaceSet(bool wantNormal, bool cleanVertices)
{
    if (_meshDirty) {
        createMesh(cleanVertices);
        _meshDirty = false;
    }

    NodeCoordinate *ncoord = (NodeCoordinate *)_scene->createNode("Coordinate");
    ncoord->point(_mesh->getVertices());
    NodeIndexedFaceSet *node = (NodeIndexedFaceSet *)
                               _scene->createNode("IndexedFaceSet");
    node->coord(new SFNode(ncoord));
    if (wantNormal) {
        NodeNormal *nnormal = NULL;
/*
        if (getType() != DUNE_NURBS_SURFACE) {
            _mesh->generateFaceNormals();
            _mesh->smoothNormals(false);
        }
*/
        if (getSmoothNormals()) {
            nnormal = (NodeNormal *)_scene->createNode("Normal");
            nnormal->vector(getSmoothNormals());
        }
        if (nnormal) {
            node->normal(new SFNode(nnormal));
            MFInt32 *ni = getSmoothNormalIndex();
            if (ni != NULL)
                node->normalIndex(ni);
        } 
    } 
    node->normalPerVertex(new SFBool(_mesh->normalPerVertex()));
    node->creaseAngle(new SFFloat(_mesh->creaseAngle())); 
    node->coordIndex(_mesh->getCoordIndex());
    node->solid(new SFBool(_mesh->solid()));
    node->ccw(new SFBool(_mesh->ccw()));
    node->convex(new SFBool(_mesh->convex()));
    NodeTextureCoordinate *ntexCoord = NULL;
    if (_mesh->getTexCoords()) {
        ntexCoord = (NodeTextureCoordinate *)
                    _scene->createNode("TextureCoordinate");
        ntexCoord->point(_mesh->getTexCoords());
    }
    if (ntexCoord) {
        node->texCoord(new SFNode(ntexCoord));
        node->texCoordIndex(_mesh->getTexCoordIndex());
    }

    return node;
}

Node * 
MeshBasedNode::toTriangleSet(void)
{
    triangulateMesh();

    NodeCoordinate *ncoord = (NodeCoordinate *)_scene->createNode("Coordinate");
    MFVec3f *vertices = _mesh->getVertices();
    MFInt32 *coordIndices = _mesh->getCoordIndex();
    MFVec3f *points = new MFVec3f();
    int numVertices = 0;
    for (int i = 0; i < coordIndices->getSize(); i++) {
         int index = coordIndices->getValue(i);
         if (index >= 0)
             points->setSFValue(numVertices++, vertices->getValue(index)); 
    }
    ncoord->point(points);
    NodeTriangleSet *node = (NodeTriangleSet *)
                            _scene->createNode("TriangleSet");
    node->coord(new SFNode(ncoord));
    node->solid(new SFBool(_mesh->solid()));
    node->ccw(new SFBool(_mesh->ccw()));
/*
    NodeTextureCoordinate *ntexCoord = NULL;
    if (_mesh->getTexCoords()) {
        ntexCoord = (NodeTextureCoordinate *)
                    _scene->createNode("TextureCoordinate");
        ntexCoord->point(_mesh->getTexCoords());
    }
    if (ntexCoord) {
        node->texCoord(new SFNode(ntexCoord));
        node->texCoordIndex(_mesh->getTexCoordIndex());
    }
*/
    return node;
}

Vec3f
MeshBasedNode::getMinBoundingBox(void) 
{ 
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }
    Vec3f ret(0, 0, 0);
    if (_mesh->getVertices())
        return _mesh->getVertices()->getMinBoundingBox();
    return ret;
}

Vec3f
MeshBasedNode::getMaxBoundingBox(void) 
{ 
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }
    Vec3f ret(0, 0, 0);
    if (_mesh->getVertices())
        return _mesh->getVertices()->getMaxBoundingBox();
    return ret;
}

MFVec3f *
MeshBasedNode::getVertices(void)
{
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }
    if (_mesh == NULL)
        return NULL;
    return _mesh->getVertices();
}

MFVec2f *
MeshBasedNode::getTextureCoordinates()
{
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }
    if (_mesh == NULL)
        return NULL;
    return _mesh->generateTextureCoordinates();
}

MFInt32 *
MeshBasedNode::getTexCoordIndex()
{
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }
    if (_mesh == NULL)
        return NULL;
    return _mesh->getCoordIndex();
}

int
MeshBasedNode::optimizeNormals(int *coordIndex, Vec3f *vertices, 
                               Vec3f *normals, int length, bool ccw,
                               bool cleanDoubleVertices)
{
    int i;
    if (length < 2) 
        return length;
    int *newCoordIndex = new int[length];
    int newIndex = 1;
    newCoordIndex[0] = coordIndex[0];
    for (i = 1; i < length; i++)
        if (coordIndex[i-1] < 0)
            newCoordIndex[newIndex++] = coordIndex[i];
        else if (coordIndex[i] < 0)
            newCoordIndex[newIndex++] = coordIndex[i];
        else {
            if ((vertices[coordIndex[i-1]] - vertices[coordIndex[i]]).length() 
                > EPSILON)
                newCoordIndex[newIndex++] = coordIndex[i];
            else {
                // reaccount normal for remaining vertex i-1
                int index1 = i-2;
                int index2 = i;
                int index3 = i+1;
                if ((index1 < 0) ||
                    (index3 > length) || 
                    (coordIndex[index1] < 0) ||
                    (coordIndex[index2] < 0) ||   
                    (coordIndex[index3] < 0)) {    
                    index1 = i-3;
                    index2 = i-2;
                    index3 = i;
                }
                if ((index1 < 0) ||
                    (index3 > length) || 
                    (coordIndex[index1] < 0) ||
                    (coordIndex[index2] < 0) ||   
                    (coordIndex[index3] < 0)) {    
                    index1 = i;
                    index2 = i+1;
                    index3 = i+2;
                }
                if ((index1 < 0) ||
                    (index3 > length) || 
                    (coordIndex[index1] < 0) ||
                    (coordIndex[index2] < 0) ||   
                    (coordIndex[index3] < 0)) {    
                    continue;
                }
                if (normals != NULL) {
                    Vec3f edge = (vertices[coordIndex[index2]] - 
                                  vertices[coordIndex[index1]]);
                    normals[coordIndex[i]] = edge.cross(
                                             vertices[coordIndex[index3]] - 
                                             vertices[coordIndex[index2]]);
                    if (!ccw) 
                        normals[coordIndex[i]] = -normals[coordIndex[i]];
                    normals[coordIndex[i]].normalize();
                    normals[coordIndex[i-1]] = normals[coordIndex[i]];
                }
            }
        }
      if (cleanDoubleVertices) {
          for (i = 0; i < newIndex; i++)
              coordIndex[i] = newCoordIndex[i];
          delete []  newCoordIndex;
      } else
          newIndex = length;
      return newIndex;
}

int
MeshBasedNode::optimizeNormals(int *coordIndex, MFVec3f *vertices, 
                               MFVec3f *normals, int length, bool ccw,
                               bool cleanDoubleVertices)
{
    int i;
    if (length < 2) 
        return length;
    int *newCoordIndex = new int[length];
    int newIndex = 1;
    newCoordIndex[0] = coordIndex[0];
    for (i = 1; i < length; i++)
        if (coordIndex[i-1] < 0)
            newCoordIndex[newIndex++] = coordIndex[i];
        else if (coordIndex[i] < 0)
            newCoordIndex[newIndex++] = coordIndex[i];
        else {
            if ((vertices->getVec(coordIndex[i-1]) - 
                 vertices->getVec(coordIndex[i])).length() > EPSILON)
                newCoordIndex[newIndex++] = coordIndex[i];
            else {
                // reaccount normal for remaining vertex i-1
                int index1 = i-2;
                int index2 = i;
                int index3 = i+1;
                if ((index1 < 0) ||
                    (index3 > length) || 
                    (coordIndex[index1] < 0) ||
                    (coordIndex[index2] < 0) ||   
                    (coordIndex[index3] < 0)) {    
                    index1 = i-3;
                    index2 = i-2;
                    index3 = i;
                }
                if ((index1 < 0) ||
                    (index3 > length) || 
                    (coordIndex[index1] < 0) ||
                    (coordIndex[index2] < 0) ||   
                    (coordIndex[index3] < 0)) {    
                    index1 = i;
                    index2 = i+1;
                    index3 = i+2;
                }
                if ((index1 < 0) ||
                    (index3 > length) || 
                    (coordIndex[index1] < 0) ||
                    (coordIndex[index2] < 0) ||   
                    (coordIndex[index3] < 0)) {    
                    continue;
                }
                if (normals != NULL) {
                    Vec3f edge1 = vertices->getVec(coordIndex[index2]) - 
                                  vertices->getVec(coordIndex[index1]);
                    Vec3f edge2 = vertices->getVec(coordIndex[index3]) - 
                                  vertices->getVec(coordIndex[index2]);
                    Vec3f normal = edge1.cross(edge2);
                    if (!ccw) 
                        normal = -normal;
                    normal.normalize();
                    normals->setVec(coordIndex[i-1], normal);
                    normals->setVec(coordIndex[i], normal);
                }
            }
        }
      if (cleanDoubleVertices) {
          for (i = 0; i < newIndex; i++)
              coordIndex[i] = newCoordIndex[i];
          delete []  newCoordIndex;
      } else
          newIndex = length;
      return newIndex;
}

int 
MeshBasedNode::countPolygons(void)
{
    int ret = 0;
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }
    MFInt32 *coordIndex = _mesh->getCoordIndex();
    if (coordIndex != NULL) {     
        int lastCoordIndex = -1;
        for (int i = 0; i < coordIndex->getSize(); i++) {
            if ((coordIndex->getValue(i) == -1) && (lastCoordIndex != -1))
                ret++;
            lastCoordIndex = coordIndex->getValue(i);
        }
        // need to count last polygon ?
        if (coordIndex->getSize() > 2)
            if (coordIndex->getValue(coordIndex->getSize() - 1) != -1)
                ret++;
    }
    return ret;
}

int 
MeshBasedNode::countPolygons2Sided(void)
{
    return isDoubleSided() ? 2 * MeshBasedNode::countPolygons() : 0;
}

int 
MeshBasedNode::countPolygons1Sided(void)
{
    return isDoubleSided() ? 0 : MeshBasedNode::countPolygons();
}

MFVec3f *
MeshBasedNode::getSmoothNormals(void)
{
    if (_mesh == NULL)
        return NULL;
    MFVec3f *v = _mesh->getNormals();
    
    if (v != NULL) {
        v =  new MFVec3f((float *)((MFVec3f *)v->copy())->getValues(), 
                         v->getSize());
    }
    return v;
}

MFInt32 *
MeshBasedNode::getSmoothNormalIndex(void)
{
    if (_mesh == NULL)
        return NULL;
    MFInt32 *i = _mesh->getNormalIndex();
    
    if (i != NULL) {
        i =  new MFInt32((int *)((MFInt32 *)i->copy())->getValues(), 
                         i->getSize());
    }
    return i;
}

void
MeshBasedNode::setTexCoordFromMesh(Node *ntexCoord)
{
    if (ntexCoord->getType() != DUNE_TEXTURE_COORDINATE)
        return;
    if (_mesh == NULL)
        return;
    MFVec2f *v = _mesh->getTexCoords();
    if (v != NULL) {
        int field = ((NodeTextureCoordinate *)ntexCoord)->point_Field();
        _scene->setField(ntexCoord, field, v);
    }
}

#define writeAc3dRef(index) { \
    int ref = _mesh->getCoordIndex()->getValue(face->getOffset() + index); \
    float texCoordX = 0;                                                   \
    float texCoordY = 0;                                                   \
    if (ref < texCoords->getSFSize()) {                                    \
        texCoordX = texCoords->getValue(ref)[0];                           \
        texCoordY = texCoords->getValue(ref)[1];                           \
    }                                                                      \
    RET_ONERROR( mywritef(f, "%d %f %f\n", ref, texCoordX, texCoordY) )    \
}

int
MeshBasedNode::writeAc3d(int f, int indent)
{
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }

    _mesh->optimize();

    Node *colorNode = NULL;
    if (getType() == DUNE_ELEVATION_GRID)
        colorNode = ((NodeElevationGrid *)this)->color()->getValue();
    else if (getType() == DUNE_GEO_ELEVATION_GRID)
        colorNode = ((NodeGeoElevationGrid *)this)->color()->getValue();
    else if (getType() == DUNE_INDEXED_FACE_SET)
        colorNode = ((NodeIndexedFaceSet *)this)->color()->getValue();
    NodeColor *color = (NodeColor *) colorNode;

    NodeMaterial *nmaterial = getMaterial();
    NodeImageTexture *nimageTexture = getImageTexture();
    NodeTextureTransform *ntextureTransfrom = getTextureTransform();

    RET_ONERROR( mywritef(f, "OBJECT poly\n") )

    if (hasName())
        RET_ONERROR( mywritef(f, "name \"%s\"\n", getName().getData()) )
        
    if (nimageTexture) {
        const char *url = nimageTexture->url()->getValue(0);
        if (url) {
            RET_ONERROR( mywritef(f, "texture %s\n", url) )
            if (ntextureTransfrom)
                RET_ONERROR( mywritef(f, "texrep %f %f\n", 
                                      ntextureTransfrom->scale()->getValue(0),
                                      ntextureTransfrom->scale()->getValue(1)) )
        }
    }
    MFFloat *colors = _mesh->getColors();
    MFInt32 *colorIndex = _mesh->getColorIndex();
    bool colorPerVertexProblem = _mesh->colorPerVertex();
    if (!colors || (colors && (colors->getSize() == 0)))
        if (colorPerVertexProblem)
            colorPerVertexProblem = false;
    if (colorPerVertexProblem) {
        const char *name = "";
        if (hasName())
            name = getName();
        else
            name = getProto()->getName(_scene->isX3d());
        fprintf(stderr, "Warning: color node of %s node can ", name);
        fprintf(stderr, "not be exported to AC3D cause of colorPerVertex TRUE");
    }
    RET_ONERROR( mywritef(f, "crease %f\n", RAD2DEG(_mesh->creaseAngle())) )
    MFVec3f *vertices = getVertices();
    RET_ONERROR( mywritef(f, "numvert %d\n", vertices->getSFSize()) )
    Matrix matrix;
    glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat *) matrix);
    int i;
    for (i = 0; i < vertices->getSFSize(); i++) {
        Vec3f v(vertices->getValue(i)[0],
                vertices->getValue(i)[1],
                vertices->getValue(i)[2]);
        v = matrix * v;
        RET_ONERROR( mywritef(f, "%f %f %f\n",v.x, v.y,v.z) )
    }
    RET_ONERROR( mywritef(f, "numsurf %d\n", _mesh->getNumFaces()) )
    int surfaceFlags = 0x10; // polygon
    surfaceFlags += _mesh->solid() ? 0: 1 << 5;
    MFVec2f* texCoords = _mesh->generateTextureCoordinates();
    for (i = 0; i < _mesh->getNumFaces(); i++) {
        RET_ONERROR( mywritef(f, "SURF 0x%x\n", surfaceFlags) )
        int index = -1; 
        bool materialToWrite = true;
        if (!colorPerVertexProblem) {
            if (colorIndex && (colorIndex->getSize() > 0)) {
                if (i < colorIndex->getSize())
                    index = colorIndex->getValue(i);
            } else
                index = i;
            int materialIndex = 0;
            if ((index >= 0) && color)
                materialIndex = color->getAc3dMaterialIndex(index);
            if ((index >= 0) && colors && (index < colors->getSize()) && 
                (materialIndex > -1)) { 
                RET_ONERROR( mywritef(f, "mat %d\n", materialIndex) )
                materialToWrite = false;
            }
        }
        if (materialToWrite)
            if (nmaterial)
                RET_ONERROR( mywritef(f, "mat %d\n", 
                                      nmaterial->getAc3dMaterialIndex()) )
            else
                RET_ONERROR( mywritef(f, "mat 0\n") )

        Face *face = _mesh->getFace(i);
        RET_ONERROR( mywritef(f, "refs %d\n", face->getNumVertices()) )
        if (_mesh->ccw())
            for (int j = 0; j < face->getNumVertices(); j++)
                writeAc3dRef(j)
        else
            for (int j = face->getNumVertices() - 1; j >= 0; j--)
                writeAc3dRef(j)
    }
    RET_ONERROR( mywritef(f, "kids 0\n") )
    _meshDirty = true;
    return 0;
}

#define UNKNOWN_MATERIAL "TOTREF"

int
MeshBasedNode::writeCattGeo(int filedes, int indent)
{ 
    char *path = _scene->getCattGeoPath();
    int len = strlen(path)+256;
    char *filename = new char[len];
    mysnprintf(filename, len - 1, "%sShape%d.geo", path,
               _scene->getAndIncCattGeoFileCounter());
    int f = open(filename, O_WRONLY | O_CREAT, 00666);
    if (f == -1)
        return -1;
    delete [] filename;

    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }

    _mesh->optimize();

    Node *parent = NULL;
    if (hasParent())
        parent = getParent();

    NodeMaterial *nmaterial = getMaterial();
   
    RET_ONERROR( mywritef(f, "OBJECT") )
    RET_ONERROR( mywritef(f, swGetWindowsLinefeed()) )

    if (hasName())
        RET_ONERROR( mywritef(f, "; Shape with vrml97 DEF name %s: ", 
                              getName().getData()) )
    RET_ONERROR( mywritef(f, swGetWindowsLinefeed()) )

    RET_ONERROR( mywritef(f, "CORNERS") )
    RET_ONERROR( mywritef(f, swGetWindowsLinefeed()) )
    MFVec3f *vertices = getVertices();
    Matrix matrix;
    glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat *) matrix);
    int cornerCounter = _scene->getCattGeoCornerCounter();
    int i;
    for (i = 0; i < vertices->getSFSize(); i++) {
        Vec3f v(vertices->getValue(i)[0], 
                vertices->getValue(i)[1], 
                vertices->getValue(i)[2]);
        v = matrix * v;        
        RET_ONERROR( mywritef(f, "%d %f %f %f%s", 
                              _scene->getAndIncCattGeoCornerCounter(),
                              v.x, v.z, v.y, swGetWindowsLinefeed()) )
    }

    RET_ONERROR( mywritef(f, "PLANES") )
    RET_ONERROR( mywritef(f, swGetWindowsLinefeed()) )
    bool meshCcw = _mesh->ccw();
    // write both sides, if solid is false
    int sides = _mesh->solid() ? 1 : 2;
    for (int side = 1; side < (sides + 1); side++) {
        if (side == 2)
                meshCcw = !meshCcw;
        for (i = 0; i < _mesh->getNumFaces(); i++) {
            Face *face = _mesh->getFace(i);
            RET_ONERROR( mywritef(f, "[%d ", 
                                  _scene->getAndIncCattGeoPlaneCounter()) )
            if (nmaterial)
                RET_ONERROR( nmaterial->writeExportMaterialName(f, 
                                                                UNKNOWN_MATERIAL
                                                                ) )
            else
                RET_ONERROR( mywritef(f, UNKNOWN_MATERIAL) )
            RET_ONERROR( mywritef(f, " / ") )
            MFInt32 *mfcoordIndex = _mesh->getCoordIndex();
            if (meshCcw)
                for (int j = face->getNumVertices() - 1; j >= 0; j--) {
                    int ref = mfcoordIndex->getValue(face->getOffset() + j);
                    RET_ONERROR( mywritef(f, "%d ", ref + cornerCounter) )
                }
            else 
                for (int j = 0; j < face->getNumVertices(); j++) {
                    int ref = mfcoordIndex->getValue(face->getOffset() + j);
                    RET_ONERROR( mywritef(f, "%d ", ref + cornerCounter) )
                }
            RET_ONERROR( mywritef(f, "/ ") )
            if (nmaterial)
                RET_ONERROR( nmaterial->writeExportMaterialName(f, 
                                                                UNKNOWN_MATERIAL
                                                               ) )
            else
                RET_ONERROR( mywritef(f, UNKNOWN_MATERIAL) )
            RET_ONERROR( mywritef(f, " ]") )
            RET_ONERROR( mywritef(f, swGetWindowsLinefeed()) )
        }            
    }

    _meshDirty = true;
    return swTruncateClose(f);
}

void
MeshBasedNode::smoothNormals(bool optimizeNormals)
{
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }
    _mesh->smoothNormals(optimizeNormals);
}

void
MeshBasedNode::flip(int index)
{
    Node *faceSet = _scene->convertToIndexedFaceSet(this);
    if (faceSet != NULL)
        faceSet->flip(index);
}

void
MeshBasedNode::swap(int fromTo)
{
    Node *faceSet = _scene->convertToIndexedFaceSet(this);
    if (faceSet != NULL)
        faceSet->swap(fromTo);
}

void            
MeshBasedNode::optimizeMesh(void)
{
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }
    _mesh->optimize();
}

void            
MeshBasedNode::triangulateMesh(void)
{
    if (_meshDirty) {
        createMesh();
        _meshDirty = false;
    }
#ifdef HAVE_GLUNEWTESS
    _mesh->triangulate();
#endif
}

MFVec3f *
MeshBasedNode::getCoordFromMesh()
{ 
    if (_mesh == NULL) {
        createMesh();
        _meshDirty = false;
    }
    return _mesh->getVertices(); 
}

MFInt32 *
MeshBasedNode::getCoordIndexFromMesh()
{ 
    if (_mesh == NULL) {
        createMesh();
        _meshDirty = false;
    }
    return _mesh->getCoordIndex(); 
}
