/*
 * MeshMorphingNode.cpp
 *
 * Copyright (C) 2005 J. "MUFTI" Scheurich
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include "MeshMorphingNode.h"
#include "Mesh.h"
#include "Field.h"
#include "Scene.h"
#include "NodeIndexedFaceSet.h"
#include "NodeCoordinateInterpolator.h"
#include "NodeNormalInterpolator.h"
#include "EventOut.h"

bool
MeshMorphingNode::writeCoordinateInterpolator(Node *node, int filedes, 
                                              int indent, bool optimize)
{
    int i;
    Array<int> fieldList;

    for (i = 0; i < _proto->getNumFields(); i++) {
        int type = _proto->getField(i)->getType();
        if (typeDefaultValue(type)->supportAnimation(_scene->isX3d()))
            fieldList.append(i);
    }
    
    // find interpolators which have routes to the morphing node
    Array<Interpolator *> interpolators;
    for (i = 0; i < fieldList.size(); i++) {
        Interpolator *inter = _scene->findUpstreamInterpolator(this, 
                                                               fieldList[i]);
        if (inter != NULL)
            interpolators.append(inter);
    }

    if (interpolators.size() == 0) {
        // nothing to do
        return true;
    }

    // get name of used input of interpolators
    // only the first found input is used
    MyString interInputName;
    MyString interInputEventName;
    bool foundInterInput = false;
    for (int k = 0; k < interpolators.size(); k++) {
        Node *node = interpolators[k];
        int eventIn = node->lookupEventIn("set_fraction", false);
        if (eventIn == INVALID_INDEX)
            continue;

        // use first eventIn found to drive Interpolator
        SocketList::Iterator *iter= node->getInput(eventIn).first();
        if (iter && iter->item().getNode()) {
            Node *src = iter->item().getNode();
            interInputName = src->getName();
            int field = iter->item().getField();
            interInputEventName = src->getProto()->getEventOut(field)->
                                       getName(src->getScene()->isX3d());
            foundInterInput = true;
            break;
        }
    }
        
    // sort list of keys
    Array<float> keys;
    Array<float> coordKeyValues;
    Array<float> normalKeyValues;
    bool foundMinKey = false;
    float currentKey = 0;
    // start list of keys with 0 
    keys.append(currentKey);
    do {
       foundMinKey = false;
       float minKey = 0;
       for (int i = 0; i < interpolators.size(); i++)
           for (int j = 0; j < interpolators[i]->getNumKeys(); j++)
               if (interpolators[i]->getKey(j) > currentKey)
                   if (foundMinKey == false) {
                       minKey = interpolators[i]->getKey(j);
                       foundMinKey = true;
                   } else if (interpolators[i]->getKey(j) < minKey)
                       minKey = interpolators[i]->getKey(j);
       if (foundMinKey) {
           keys.append(minKey);
           currentKey = minKey;
       }
    } while (foundMinKey);


    // get data from Interpolators and generate mesh

    void *data = initializeData();
    if (data == NULL)
        return true;
    for (i = 0; i < keys.size(); i++) {
        for (int j = 0; j < fieldList.size(); j++) {
            int field = fieldList[j];
            Interpolator *inter = _scene->findUpstreamInterpolator(this, field);
            if (inter != NULL)
                loadDataFromInterpolators(data, inter, field, keys[i]);
        }

        createMeshFromData(data, optimize);

        // check if coordIndex of optimized mesh is identical to optimized node
        if (optimize) {
            MFInt32 *micoordIndex = ((NodeIndexedFaceSet *)node)->coordIndex();
            if (!micoordIndex->equals(_mesh->getCoordIndex()))
                return false;
        }

        MFVec3f *vertices = _mesh->getVertices();
        for (int k = 0; k < vertices->getSize(); k++)
            coordKeyValues.append(vertices->getValues()[k]);
        if (TheApp->GetNormalsOnPureVRML97()) {
            MFVec3f *normals = getSmoothNormals();
            for (int k = 0; k < normals->getSize(); k++)
                normalKeyValues.append(normals->getValues()[k]);
        }
    }
    finalizeData(data);
    data = NULL;    


    NodeCoordinateInterpolator *coordInter = (NodeCoordinateInterpolator *) 
          _scene->createNode("CoordinateInterpolator");

    // write CoordinateInterpolator
    MFFloat *mfkeys = new MFFloat((float *)keys.getData(), keys.size());      
    coordInter->key((MFFloat *)mfkeys->copy());

    MFVec3f *mfCoordKeyValues = new MFVec3f((float *)coordKeyValues.getData(),
                                             coordKeyValues.size());
    coordInter->keyValue((MFVec3f *)mfCoordKeyValues->copy());
    _scene->addDelayedWriteNode(coordInter); 

    // write route
    MyString routeString = "ROUTE ";
    routeString += coordInter->getName(true);
    routeString += ".value_changed TO ";
    Node *ncoord = ((NodeIndexedFaceSet *)node)->coord()->getValue();
    routeString += ncoord->getName(true);
    routeString += ".set_point";
    _scene->addRouteString(routeString);

    if (foundInterInput) {
        routeString = "ROUTE ";
        routeString += interInputName;
        routeString += ".";
        routeString += interInputEventName;
        routeString += " TO ";
        routeString += coordInter->getName(true);
        routeString += ".set_fraction";
        _scene->addEndRouteString(routeString);
    }

    if (TheApp->GetNormalsOnPureVRML97()) {
        // write NormalInterpolator
        NodeNormalInterpolator *normalInter = (NodeNormalInterpolator *) 
               _scene->createNode("NormalInterpolator");
    
        normalInter->key((MFFloat *)mfkeys->copy());

        MFVec3f *mfNormalKeyValues = new MFVec3f((float *)
                                                 normalKeyValues.getData(),
                                                 normalKeyValues.size());
        normalInter->keyValue((MFVec3f *)mfNormalKeyValues->copy());
        _scene->addDelayedWriteNode(normalInter); 

        // write Route
        MyString routeString = "ROUTE ";
        routeString += normalInter->getName(true);
        routeString += ".value_changed TO ";
        Node *nnormal = ((NodeIndexedFaceSet *)node)->normal()->getValue();
        routeString += nnormal->getName(true);
        routeString += ".set_vector";
        _scene->addRouteString(routeString);

        if (foundInterInput) {
            routeString = "ROUTE ";
            routeString += interInputName;
            routeString += ".";
            routeString += interInputEventName;
            routeString += " TO ";
            routeString += normalInter->getName(true);
            routeString += ".set_fraction";
            _scene->addEndRouteString(routeString);
        }
    }   
    return true;
} 

