/*
 * Node.cpp
 *
 * Copyright (C) 1999 Stephen F. White, 2005 J. "MUFTI" Scheurich
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "stdafx.h"

#include "Node.h"
#include "Scene.h"
#include "Field.h"
#include "FieldValue.h"
#include "FieldCommand.h"
#include "Proto.h"
#include "Path.h"
#include "SFNode.h"
#include "MFNode.h"
#include "SFString.h"
#include "MFString.h"
#include "EventIn.h"
#include "EventOut.h"
#include "ExposedField.h"
#include "URL.h"
#include "DuneApp.h"
#include "NodeScript.h"

#define HANDLE_SIZE 2.0f

NodeData::NodeData(Scene *scene, Proto *proto)
{
    static unsigned long identifier_source=0;
    identifier_source++;
    _identifier=identifier_source;

    _scene = scene;
    _graphX = _graphY = 0;
    _graphWidth = _graphHeight = 0;
    _refs = 0;
    _flags = 1<<NODE_FLAG_COLLAPSED;
    _proto = proto;

    int i;
    _numFields = _proto->getNumFields();
    _fields = new FieldValue *[_numFields];
    for (i = 0; i < _numFields; i++) {
        _fields[i] = _proto->getField(i)->getDefault()->copy();
        _fields[i]->ref();
    }
    _numEventIns = _proto->getNumEventIns();
    _inputs = new SocketList[_numEventIns];
    _numEventOuts = _proto->getNumEventOuts();
    _outputs = new SocketList[_numEventOuts];
    _numberUse = 1;
    _inSideProto = false;
    _nodePROTO = NULL;
    _isMesh = false;
    handleIs();
    generateTreeLabel();
    _scene->addNode((Node*)this);
}

NodeData::NodeData(const Node &node)
{
    _identifier=node._identifier;  // FIXME: unclear: should all IDs be unique??

    _scene = node._scene;
    _refs = 0;
    _graphX = _graphY = 0;
    _graphWidth = _graphHeight = 0;
    _flags = node._flags;
    _proto = node._proto;
    _numFields = _proto->getNumFields();
    _fields = new FieldValue *[_numFields];
    if (node._name.length() != 0)
        if (!_scene->hasAlreadyName(node._name)) 
            _scene->def(node._name, (Node *)this);
        else
            _scene->makeSimilarName((Node *)this, node._name);
    generateTreeLabel();        
    for (int i = 0; i < _numFields; i++) {
        _fields[i] = node._fields[i]->copy();
        _fields[i]->ref();
    }
    _numEventIns = _proto->getNumEventIns();
    _inputs = new SocketList[_numEventIns];
    _numEventOuts = _proto->getNumEventOuts();
    _outputs = new SocketList[_numEventOuts];
    _numberUse = 1;
    _inSideProto = node._inSideProto;
    _nodePROTO = node._nodePROTO;
    _parents = node._parents;
    handleIs();
    _scene->addNode((Node*)this);
}

NodeData::~NodeData()
{
    unref();
}

void NodeData::delete_this(void)
{
    _scene->removeNode((Node *)this);
    for (int i = 0; i < _numFields; i++) {
        _fields[i]->unref();
    }
    _numFields = 0;
    delete [] _fields;
    delete [] _inputs;
    delete [] _outputs;
    _scene->undef(_name);
}

Node::Node(Scene *scene, Proto *proto) : NodeData(scene,proto) 
{
    _commentsList = new NodeList;
    _geometricParentIter = NULL;     
    ref();
}

Node::Node(const Node &node) : NodeData(node)
{
    _geometricParentIter = node._geometricParentIter;
    _commentsList = new NodeList;
    ref();
}

Node::Node(const Node &node, Proto *proto) : NodeData(node)
{ 
    _geometricParentIter = node._geometricParentIter;
    _proto = proto;
    for (int i = 0; i < proto->getNumEventOuts(); i++)
        _outputs[i] = node._outputs[i];
    _commentsList = new NodeList;
    ref();
}

Node::~Node()
{
    delete _commentsList;
}

bool
NodeData::hasName(void)
{
    if (_name[0]) 
        return true;
    else
        return false;
}

const MyString& 
NodeData::getName(bool newName)
{
    if (newName || needsDEF())
       if (_name.length() == 0) 
          _scene->generateUniqueNodeName((Node *)this);       
    return _name; 
}

void
NodeData::setName(const char *name)
{
    _scene->undef(_name);
    _name = name; 
    generateTreeLabel();
}

bool
NodeData::needsDEF() const
{
    if (_name.length() != 0) return true;

    if (((Node *)this)->getNumParents() > 1) return true;

    for (int i = 0; i < _numEventIns; i++) {
        if (_inputs[i].size() > 0) {
            return true;
        }
    }

    for (int j = 0; j < _numEventOuts; j++) {
        if (_outputs[j].size() > 0) {
            return true;
        }
    }

    return false;
}

FieldValue *
NodeData::getField(int index) const
{
    assert(index >= 0 && index < _numFields);
    return _fields[index];
}

void
NodeData::setField(int index, FieldValue *value)
{
    assert(index >= 0 && index < _numFields);
    if (_fields[index] != NULL)
        assert(_fields[index]->getType() == value->getType());

    // if field is an SFNode or MFNode type, remove old values from 
    // children's parent list

    if (value->getType() == SFNODE) {
        Node *child = ((SFNode *) _fields[index])->getValue();
        if (child) child->removeParent();
    } else if (value->getType() == MFNODE) {
        NodeList *childList = ((MFNode *) _fields[index])->getValues();
        if (childList) {
            for (int i = 0; i < childList->size(); i++) {
                Node *child = childList->get(i);
                if (child) child->removeParent();
            }
        }
    }

    value->clamp(_proto->getField(index)->getMin(), 
                 _proto->getField(index)->getMax());

    _fields[index]->unref();
    _fields[index] = value;
    _fields[index]->ref();

    if (value->getType() == SFNODE) {
        Node *child = ((SFNode *) value)->getValue();
        if (child) child->addParent((Node*)this, index);
    } else if (value->getType() == MFNODE) {
        NodeList *childList = ((MFNode *) value)->getValues();
        if (childList) {
            for (int i = 0; i < childList->size(); i++) {
                Node *child = childList->get(i);
                if (child) child->addParent((Node*)this, index);
            }
        }
    }
    // handle IS
    Field *field = _proto->getField(index);
    if (field && (field->getFlags() & FF_IS))
        for (int i = 0; i < field->getNumIs(); i++) {
            if (field->getIsNodeIndex(i) >= 0) {
                Node *isNode = getIsNode(field->getIsNodeIndex(i));
                isNode->setField(field->getIsField(i), value);
            }
        }
    ExposedField *expField = field->getExposedField();
    if (expField && (expField->getFlags() & FF_IS))
        for (int i = 0; i < expField->getNumIs(); i++) {
            if (expField->getIsNodeIndex(i) >= 0) {
                Node *isNode = getIsNode(expField->getIsNodeIndex(i));
                isNode->setField(expField->getIsField(i), value);
            }
        }
}

void
Node::addFieldNodeList(int index, NodeList *childList)
{
    assert(index >= 0 && index < _numFields);
    if (((MFNode*)_fields[index])->getSize() == 0)
        setField(index, new MFNode(childList));
    else {
        for (int i = 0; i < childList->size(); i++) {
            Node *child = childList->get(i);
            FieldValue* newField= ((MFNode*)_fields[index])->addNode(child);
            newField->clamp(_proto->getField(index)->getMin(), 
                            _proto->getField(index)->getMax());

            _fields[index]->unref();
            _fields[index] = newField;
            _fields[index]->ref();
            if (child) {
                child->addParent(this, index);
            }
        }
    }
    ((MFNode*)_fields[index])->getValues();
}

int NodeData::write(int f, int indent)
{
    bool x3d = _scene->isX3dv();
    if (_proto) {
        TheApp->checkSelectionLinenumberCounting(_scene, (Node*) this);
        if (getFlag(NODE_FLAG_DEFED)) {
            RET_ONERROR( mywritestr(f, "USE ") )
            RET_ONERROR( mywritestr(f, (const char *) _name) )
            RET_ONERROR( mywritestr(f, "\n") )
            TheApp->incSelectionLinenumber();
        } else {
            if (((Node *)this)->needsDEF()) {
                if (!_name[0]) _scene->generateUniqueNodeName((Node*)this);
                RET_ONERROR( mywritestr(f, "DEF ") )
                RET_ONERROR( mywritestr(f, (const char *) _name) )
                RET_ONERROR( mywritestr(f, " ") )
            }
            setFlag(NODE_FLAG_DEFED);
            bool writeProtoName = true;
            if (TheApp->getCoverMode() && !hasDefault(FF_COVER_ONLY)) {
                Proto *extensionProto = _scene->getExtensionProto(_proto);
                if (extensionProto != NULL) {
                    writeProtoName = false;
                    RET_ONERROR( mywritestr(f, extensionProto->getName(x3d)) )
                }
            }
            if (TheApp->getKambiMode() && !hasDefault(FF_KAMBI_ONLY)) {
                Proto *extensionProto = _scene->getExtensionProto(_proto);
                if (extensionProto != NULL) {
                    writeProtoName = false;
                    RET_ONERROR( mywritestr(f, extensionProto->getName(x3d)) )
                }
            }
            if (writeProtoName)
                RET_ONERROR( mywritestr(f, (const char *) _proto->getName(x3d)))
            RET_ONERROR( mywritestr(f, " ") )
            if (!TheApp->GetkrFormating()) {
                RET_ONERROR( mywritestr(f, "\n") )
                TheApp->incSelectionLinenumber();
                RET_ONERROR( indentf(f, indent + TheApp->GetIndent()) )
            }
            RET_ONERROR( mywritestr(f, "{\n") )
            TheApp->incSelectionLinenumber();
            RET_ONERROR( writeFields(f, indent + TheApp->GetIndent()) )
            if (!TheApp->GetkrFormating())
                RET_ONERROR( indentf(f, indent + TheApp->GetIndent()) )
            else
                RET_ONERROR( indentf(f, indent) )
            RET_ONERROR( mywritestr(f, "}\n") )
            TheApp->incSelectionLinenumber();
            if (indent==0) {
                RET_ONERROR( mywritestr(f, "\n") )
                TheApp->incSelectionLinenumber();
            }
            RET_ONERROR( writeRoutes(f, indent) )
            setFlag(NODE_FLAG_TOUCHED);
        }
    }
    return(0);
}

int            
NodeData::writeProto(int filedes)
{
    // used to write the EXTERNPROTO definition of extension nodes
    return 0;
}

int
NodeData::writeProto(int f, const char *urn, const char *directoryName,
                           const char *haveUrl)
{
    const char* nodeName = _proto->getName(_scene->isX3d());
    RET_ONERROR( mywritef(f, "EXTERNPROTO %s[\n", nodeName) )    
    TheApp->incSelectionLinenumber();
    RET_ONERROR( writeProtoArguments(f) )
    RET_ONERROR( mywritestr(f, " ]\n") )
    TheApp->incSelectionLinenumber();
    RET_ONERROR( mywritestr(f, "[\n") )
    TheApp->incSelectionLinenumber();
    if (strlen(urn) > 0) {
        RET_ONERROR( mywritef(f, "  \"%s:%s\"\n", urn, nodeName) )
        TheApp->incSelectionLinenumber();
    }
    RET_ONERROR( mywritef(f, "  \"%sPROTO.wrl\"\n", nodeName) )
    TheApp->incSelectionLinenumber();
    RET_ONERROR( mywritestr(f, "  \"") )
    char *dunedocs = getenv("DUNEDOCS");
    if (dunedocs != NULL) {
        RET_ONERROR( mywritestr(f, dunedocs) )
        RET_ONERROR( mywritestr(f, "/") )
        RET_ONERROR( mywritestr(f, directoryName) )
        RET_ONERROR( mywritestr(f, "/") )
    } else if (strlen(haveUrl) > 0) {
        RET_ONERROR( mywritestr(f, haveUrl) )
        RET_ONERROR( mywritestr(f, "/") )
    }
    RET_ONERROR( mywritef(f, "%sPROTO.wrl\"\n", nodeName) )
    TheApp->incSelectionLinenumber();
    RET_ONERROR( mywritef(f, 
                          "  \"%s/docs/%s/%sPROTO.wrl\"\n",
                          "http://129.69.35.12/dune",
                          directoryName, nodeName) )
    TheApp->incSelectionLinenumber();
    RET_ONERROR( mywritestr(f, "]\n") )
    TheApp->incSelectionLinenumber();
    return 0;
}

int 
NodeData::writeIs(int f, int indent, const char *name, const char* isName)
{
    RET_ONERROR( indentf(f, indent) )
    RET_ONERROR( mywritestr(f, name) )
    RET_ONERROR( mywritestr(f, " IS ") )
    RET_ONERROR( mywritestr(f, isName) )
    RET_ONERROR( mywritestr(f, "\n") )
    TheApp->incSelectionLinenumber();
    return 0;
}

int NodeData::writeFields(int f, int indent)
{
    int i;

    if (!_proto) return(0);

    for (i = 0; i < _numFields; i++)
        RET_ONERROR( writeField(f, indent, i) )

    for (i = 0; i < _numEventIns; i++)
        RET_ONERROR( writeEventIn(f, indent, i) )

    for (i = 0; i < _numEventOuts; i++)
        RET_ONERROR( writeEventOut(f, indent, i) )

    return(0);
}

int NodeData::writeField(int f, int indent,int i, bool script)
{
    if (!_proto) return(0);
    const char *oldBase = _scene->getURL();
    const char *newBase = _scene->getNewURL();
    bool        tempSave = _scene->isTempSave();
    bool        x3d = _scene->isX3dv();

    Field *field = _proto->getField(i);
    if (x3d && (field->getFlags() & FF_VRML_ONLY))
        return 0;
    if (!x3d && (field->getFlags() & FF_X3D_ONLY))
        return 0;
    FieldValue *value = _fields[i];
    const char *name = field->getName(x3d);
    const char *isName = NULL;
    if (isInsideProto())
        for (int j = 0; j < _scene->getNumProtoNames(); j++) {
            const char *protoName = _scene->getProtoName(j);
            Proto *proto = _scene->getProto(protoName);
            int isField = proto->lookupIsField((Node *)this, i);
            if (isField != -1) {
                isName = proto->getField(isField)->getName(x3d);
                break;
            }
            int isExposedField = proto->lookupIsExposedField((Node *)this, i);
            if (isExposedField != -1) {
                isName = proto->getExposedField(isExposedField)->getName(x3d);
                break;
            }
        }
    if (isName)
        RET_ONERROR( writeIs(f, indent, name, isName) )
    else if (value) {
        if (script || !value->equals(field->getDefault())) {                
            RET_ONERROR( indentf(f, indent) )
            RET_ONERROR( mywritestr(f , name) )
            RET_ONERROR( mywritestr(f ," ") )
            if ((field->getFlags() & FF_URL) && (!TheApp->GetKeepURLs())) {
                value = rewriteField(value, oldBase, newBase);
                RET_ONERROR( value->write(f, indent) )
                if (!tempSave) {
                     setField(i, value);
                     FieldUpdate* fieldUpdate=new FieldUpdate((Node*)this, i);
                     _scene->UpdateViews(NULL, UPDATE_FIELD,
                                         (Hint *) fieldUpdate);    
                     delete fieldUpdate;
                } else {
                    delete value;
                }
            } else {
                RET_ONERROR( value->write(f, indent) )
            }
        }
    }
    return(0);
}

int NodeData::writeEventIn(int f, int indent, int i, bool eventName)
{
    if (!_proto) return(0);

    bool x3d = _scene->isX3dv();

    EventIn *eventIn = _proto->getEventIn(i);
    const char *name = eventIn->getName(x3d);
    const char *isName = NULL;
    if (isInsideProto())
        for (int j = 0; j < _scene->getNumProtoNames(); j++) {
            const char *protoName = _scene->getProtoName(j);
            Proto *proto = _scene->getProto(protoName);
            int isEventIn = -1;
            ExposedField *exposedField = eventIn->getExposedField();
            int field = i;
            int type = EL_EVENT_IN;
            if (exposedField) {
                field = exposedField->getField();
                type = EL_EXPOSED_FIELD;
            }
            isEventIn = proto->lookupIsEventIn((Node *)this, field, type);
            if (isEventIn != -1) {
                isName = proto->getEventIn(isEventIn)->getName(x3d);
                break;
            }
        if (isName != NULL)
            break;
        }
    if (isName)
         RET_ONERROR( writeIs(f, indent, name, isName) )
    else if (eventName) {
        RET_ONERROR( mywritestr(f, name) )
        RET_ONERROR( mywritestr(f, "\n") )
        TheApp->incSelectionLinenumber();
    }
    return(0);
}

int NodeData::writeEventOut(int f, int indent, int i, bool eventName)
{
    if (!_proto) return(0);

    bool x3d = _scene->isX3dv();

    EventOut *eventOut = _proto->getEventOut(i);
    const char *name = eventOut->getName(x3d);
    const char *isName = NULL;
    if (isInsideProto())
        for (int j = 0; j < _scene->getNumProtoNames(); j++) {
            const char *protoName = _scene->getProtoName(j);
            Proto *proto = _scene->getProto(protoName);
            int isEventOut = -1;
            ExposedField *exposedField = eventOut->getExposedField();
            int field = i;
            int type = EL_EVENT_OUT;
            if (exposedField) {
                field = exposedField->getField();
                type = EL_EXPOSED_FIELD;
            }
            isEventOut = proto->lookupIsEventOut((Node *)this, field, type);
            if (isEventOut != -1) {
                isName = proto->getEventOut(isEventOut)->getName(x3d);
                break;
            }
        if (isName != NULL)
            break;
        }
    if (isName)
         RET_ONERROR( writeIs(f, indent, name, isName) )
    else if (eventName) {
        RET_ONERROR( mywritestr(f, name) )
        RET_ONERROR( mywritestr(f, "\n") )
        TheApp->incSelectionLinenumber();
    }
    return(0);
}

int NodeData::writeRoutes(int f, int indent) const
{
    if (!_proto) return(0);

    bool x3d = _scene->isX3dv();

    int i;
    SocketList::Iterator *j;

    for (i = 0; i < _numEventIns; i++) {
        for (j = _inputs[i].first(); j != NULL; j = j->next()) {
            if (j->item().getNode()->getFlag(NODE_FLAG_TOUCHED)) {
                MyString routestring;

                Node *src = j->item().getNode();
                if (_scene->isPureVRML() &&  
                    ((getNodeClass() & PARAMETRIC_GEOMETRY_NODE) || 
                     (src->getNodeClass() & PARAMETRIC_GEOMETRY_NODE)))
                    continue;
                routestring="ROUTE ";
                routestring+=src->getName(x3d);
                routestring+=".";
                routestring+=src->getProto()->
                             getEventOut(j->item().getField())->getName(x3d);
                routestring+=" TO ";
                routestring+=_name;
                routestring+=".";
                routestring+=getProto()->getEventIn(i)->getName(x3d);
                _scene->addRouteString(routestring);
            }
        }
    }

    for (i = 0; i < _numEventOuts; i++) {
        for (j = _outputs[i].first(); j != NULL; j = j->next()) {
            if (j->item().getNode()->getFlag(NODE_FLAG_TOUCHED)) {
                MyString routestring;

                Node *dst = j->item().getNode();
                if (_scene->isPureVRML() &&  
                    ((getNodeClass() & PARAMETRIC_GEOMETRY_NODE) || 
                     (dst->getNodeClass() & PARAMETRIC_GEOMETRY_NODE)))
                    continue;
                routestring="ROUTE ";
                routestring+=_name;
                routestring+=".";
                routestring+=getProto()->getEventOut(i)->getName(x3d);
                routestring+=" TO ";
                routestring+=dst->getName(x3d);
                routestring+=".";
                routestring+=dst->getProto()->getEventIn(j->item().getField())->
                             getName(x3d);
                _scene->addRouteString(routestring);
            }
        }
    }

#ifndef HAVE_ROUTE_AT_END
    if (indent==0)
       RET_ONERROR( _scene->writeRouteStrings(f, indent) )
#endif
    return(0);
}

int
NodeData::lookupEventIn(const MyString &name, bool x3d) const
{
    return _proto ? _proto->lookupEventIn(name, x3d) : INVALID_INDEX;
}

int
NodeData::lookupEventOut(const MyString &name, bool x3d) const
{
    return _proto ? _proto->lookupEventOut(name, x3d) : INVALID_INDEX;
}

MyString
NodeData::newEventName(int typeEnum, bool out)
{
    bool x3d = _scene->isX3d();
    char name[1024];
    MyString eventName = "";
    const char *typestr = typeEnumToString(typeEnum);
    if (typestr[0] == 'M')
        eventName += "m";
    for (int i = 2; i < strlen(typestr); i++)
        eventName += tolower(typestr[i]);
    int counter = 0;
    bool foundflag;
    do {
        foundflag = false;
        counter++;
        const char *format = "%d_in";
        if (out)
            format = "%d_out";
        mysnprintf(name, 1023, (const char *)format, counter);

        int numberEvents = 0;
        if (out)
            numberEvents = _proto->getNumEventOuts();
        else
            numberEvents = _proto->getNumEventIns();
        for (int i = 0; i < numberEvents; i++) {
            const char *oldName;
            if (out)
               oldName = (const char*) _proto->getEventOut(i)->getName(x3d);
            else 
               oldName = (const char*) _proto->getEventIn(i)->getName(x3d);
            MyString cmpName = "";
            cmpName += eventName;
            cmpName += name; 
            if (strcmp(oldName, (const char*)cmpName) == 0)
                foundflag = true;
        }
    } while (foundflag == true);
    eventName += name;
    return eventName;
}

void
NodeData::addInput(int eventIn, Node *src, int eventOut)
{
    if (eventIn >= _numEventIns)
        if (getType() == DUNE_SCRIPT) {
            int typeEnum = src->getProto()->getEventOut(eventOut)->getType();
            ((NodeScript *)this)->addEventIn(typeEnum, 
                                             newEventName(typeEnum, false));
            update();
            NodeUpdate *hint= new NodeUpdate((Node *)this, NULL, 0);
            _scene->UpdateViews(NULL, UPDATE_CHANGE_INTERFACE_NODE, 
                                (Hint*) hint);
        }
    _inputs[eventIn].append(RouteSocket(src, eventOut));
}

void
NodeData::addOutput(int eventOut, Node *dst, int eventIn)
{
    if (eventOut >= _numEventOuts)
        if (getType() == DUNE_SCRIPT) {
            int typeEnum = dst->getProto()->getEventIn(eventIn)->getType();
            ((NodeScript *)this)->addEventOut(typeEnum, 
                                              newEventName(typeEnum, true));
            update();
            NodeUpdate *hint= new NodeUpdate((Node *)this, NULL, 0);
            _scene->UpdateViews(NULL, UPDATE_CHANGE_INTERFACE_NODE, 
                                (Hint*) hint);
        }
    _outputs[eventOut].append(RouteSocket(dst, eventIn));
}

void
NodeData::removeInput(int eventIn, Node *src, int eventOut)
{
    SocketList::Iterator *j = _inputs[eventIn].find(RouteSocket(src, eventOut));
    if (j) 
        _inputs[eventIn].remove(j);
}

void
NodeData::removeOutput(int eventOut, Node *dst, int eventIn)
{
    SocketList::Iterator *j = _outputs[eventOut].find(RouteSocket(dst, 
                                                                  eventIn));
    if (j)
        _outputs[eventOut].remove(j);
}

void
NodeData::update()
{
    // used by Script to update its fields and
    // used by shapes to redraw after a change of a MF-field
}

void
NodeData::reInit()
{
    // used to reinitialise private data of a "copy"-ed node
    // e.g. to reinitialise mesh data of some geometric shapes
}


void
Node::addParent(Node *parent, int field)
{
    _parents.append(Parent(parent, field, this));
    _geometricParentIter = _parents.last();
}

int
NodeData::findChild(Node *child, int field) const
{
    assert(field >= 0 && field < _numFields);

    FieldValue *value = _fields[field];

    if (value->getType() == SFNODE) {
        return (((SFNode *) value)->getValue() == child) ? 0 : -1;
    } else if (value->getType() == MFNODE) {
        NodeList    *list = ((MFNode *) value)->getValues();
        for (int i = 0; i < list->size(); i++) {
            if (list->get(i) == child) return i;
        }
        return -1;
    } else {
        return -1;
    }
}

bool
Node::hasAncestor(Node *node) const
{
    if (this != _scene->getRoot()) {
        if (hasParent()) {
            Node *parent = getParent();
            if (parent == node || parent->hasAncestor(node)) 
                return true;
        }  
    }
    return false;
}

bool 
NodeData::checkValidX3d(Element *field)
{
    if (_scene != NULL)
        if (_scene->isX3d()) {
            if (field->getFlags() & FF_VRML_ONLY)
                return false;
            else if (field->getFlags() & FF_X3D_ONLY)
                return false;
        } else {
            if (field->getFlags() & FF_X3D_ONLY)
                return false;
            else if (field->getFlags() & FF_VRML_ONLY)
                return false;
        }
    return true;
}

int
NodeData::findValidFieldType(int childType)
{
    int match = -1;
    for (int field = 0; field < _numFields; field++) {
        if (validChildType(field, childType)) {
            if (match != -1) {
                // ambiguous
                return - 1; 
            }
            match = field;
        }
    }
    return match;
}

int
NodeData::findFirstValidFieldType(int childType)
{
    for (int field = 0; field < _numFields; field++) {
        if (validChildType(field, childType)) {
            return field;
        }
    }
    return -1;
}

int
NodeData::findValidField(Node *child)
{
    int ret = findValidFieldType(child->getNodeClass());
    if (ret == -1)
        ret = findValidFieldType(child->getType());
    return ret;
}

bool
NodeData::validChild(int field, Node *child)
{
    return validChildType(field, child->getNodeClass());
}

bool
NodeData::validChildType(int field, int childType)
{
    if (field < 0 || field >= _numFields) return false;

    Field *def = _proto->getField(field);

    // SFNode field already in use ?
    if (def->getType() == SFNODE && 
        ((SFNode *) _fields[field])->getValue() != NULL)
        return false;

    if ((def->getType() == SFNODE) || (def->getType() == MFNODE)) {
        int nodeType = def->getNodeType();

        if (nodeType == ANY_NODE) {
            return checkValidX3d(def);
        } else if (childType == nodeType) {
            return checkValidX3d(def);
        } else if (isClassType(childType) && isClassType(nodeType) &&  
                   (childType & nodeType)) {
            return checkValidX3d(def);
        } else
           return false;
    } else {
        return false;
    }
}

void
NodeData::setFlagRec(int flag)
{
    setFlag(flag);

    for (int i = 0; i < _numFields; i++) {
        if (_fields[i]->getType() == MFNODE) {
            ((MFNode *) _fields[i])->getValues()->setFlag(flag);
        } else if (_fields[i]->getType() == SFNODE) {
            SFNode *value = (SFNode *) _fields[i];
            if (value->getValue()) value->getValue()->setFlagRec(flag);
        }
    }
}

void
NodeData::clearFlagRec(int flag)
{
    setFlag(flag);

    for (int i = 0; i < _numFields; i++) {
        if (_fields[i]->getType() == MFNODE) {
            ((MFNode *) _fields[i])->getValues()->clearFlag(flag);
        } else if (_fields[i]->getType() == SFNODE) {
            SFNode *value = (SFNode *) _fields[i];
            if (value->getValue()) value->getValue()->clearFlagRec(flag);
        }
    }
}

Path *
NodeData::getPath() const
{

    int len = 0;
    const Node *node, *root = _scene->getRoot();

    for (node = (Node*)this; node != root; node = node->getParent()) {
        len += 2;
    }

    int *list = new int[len];

    int i = len - 1;
    
    for (node = (Node*)this; node != root; node = node->getParent()) {
        Node *parent = node->getParent();
        int field = node->getParentField();
        FieldValue *value = parent->getField(field);
        if (value->getType() == MFNODE) {
            list[i--] = ((MFNode *) value)->getValues()->find((Node *) node);
        } else if (value->getType() == SFNODE)  {
            list[i--] = 0;
        } else {
            assert(0);
        }
        list[i--] = field;
    }

    Path    *path = new Path(list, len, _scene);
    delete [] list;
    return path;
}


bool
Node::isInScene(Scene* scene) const
{
    if (((NodeData *)this)->isInsideProto()) {
         Node *node = ((NodeData *)this)->getNodePROTO();
         if (node)
             return node->isInScene(scene);
         else
             return false;
    }
    if (scene != _scene) 
       return false;
    if (this == _scene->getRoot()) return true;
    if (!this->hasParent())
       return false;
    if (this->getParent()->isInScene(scene))
       return true;
    return false;
}

void
NodeData::sendEvent(int eventOut, double timestamp, FieldValue *value)
{
    assert(eventOut >= 0 && eventOut <= _numEventOuts);

    // handle IS
    bool noIs = true;
    EventOut *evOut = _proto->getEventOut(eventOut);
    if (evOut && (evOut->getFlags() & FF_IS))
        for (int i = 0; i < evOut->getNumIs(); i++)
            if (evOut->getFlags() & EOF_IS_HIDDEN) {
                noIs = false;
                Node *isNode = getIsNode(evOut->getIsNodeIndex(i));
                isNode->sendEvent(evOut->getIsField(i), timestamp, value);
            }

    if (noIs) {

        value->ref();

        SocketList::Iterator    *i;

        for (i = _outputs[eventOut].first(); i != NULL; i = i->next()) {
            RouteSocket s = i->item();
            s.getNode()->receiveEvent(s.getField(), timestamp, value);
        }

        value->unref();
    }
}

void
NodeData::receiveEvent(int eventIn, double timestamp, FieldValue *value)
{
    // check to see if this eventIn is part of an exposedField or
    // conntected to a normal field

    int field = -1;
    ExposedField *e = _proto->getEventIn(eventIn)->getExposedField();

    if (e) {
        field = e->getField();
        eventIn = e->getEventIn();
    } else
        field = _proto->getEventIn(eventIn)->getField();

    // handle IS
    EventIn *evIn = _proto->getEventIn(eventIn);
    if (evIn && (evIn->getFlags() & FF_IS))
        for (int i = 0; i < evIn->getNumIs(); i++) {
            Node *isNode = getIsNode(evIn->getIsNodeIndex(i));
            isNode->receiveEvent(evIn->getIsField(i), timestamp, value);
        }
    else {
        if (field != -1) {
            // set the appropriate field
            setField(field, value);
            _scene->OnFieldChange((Node*)this, field);

        // fire off an event here?
        }
    }        
}

/// compare content
bool
NodeData::isEqual(Node* node)
{
    if (_identifier==node->_identifier)
       return true;
    else
       return false;
}

bool
NodeData::hasFieldFlag(int flag)
{
    for (int field = 0; field < _numFields; field++) {
        Field *def = _proto->getField(field);
        if (def->getFlags() & FF_URL)
            return true;
    }
    return false;
}

bool
hasRoute(SocketList socketlist)
{
   if (socketlist.first() == NULL) 
      return false;
   else
      return true;
}

bool 
Node::supportAnimation(void)
{
    Proto *proto = getProto();
    for (int i = 0; i < proto->getNumEventIns(); i++) {
        int type = proto->getEventIn(i)->getType();
        if (typeDefaultValue(type)->supportAnimation(_scene->isX3d()))
            return true;
    }
    return false;
}


bool 
Node::supportInteraction(void)
{
    Proto *proto = getProto();
    for (int i = 0; i < proto->getNumEventIns(); i++) {
        int type = proto->getEventIn(i)->getType();
        if (typeDefaultValue(type)->supportInteraction())
            return true;
    }
    return false;
}

bool
Node::isInvalidChild(void)
{
    if (this == _scene->getRoot())
        return true;
    if (isInvalidChildNode() &&
        (getParent()->getProto()->getField(getParentField())->getType() 
         == MFNODE)
       )
        return true;
    return false;
}

void
Node::removeParent(void)
{
    if (_geometricParentIter != NULL)
        if (_parents.findIterator(_geometricParentIter)) {
            _parents.remove(_geometricParentIter);
            _geometricParentIter = NULL;
        }
     
}

Vec3f               
Node::getMinBoundingBox(void)
{
   Vec3f ret(0, 0, 0);
   return ret;
}

Vec3f               
Node::getMaxBoundingBox(void)
{
   Vec3f ret(0, 0, 0);
   return ret;
}

void 
Node::appendComment(Node *node)
{ 
    _commentsList->append(node); 
}

void Node::appendTo(NodeList* nodelist)
{
    for (int i = 0;i < _commentsList->size(); i++)
       nodelist->append((*_commentsList)[i]);
    nodelist->append(this);
    _commentsList->resize(0);
}

void                
Node::doWithParents(DoWithNodeCallback callback, void *data,
                    bool searchInRest, bool callSelf)
{
    for (ParentIter i = _parents.first(); i != NULL; i = i->next()) {
        bool call = true;
        if (i == _geometricParentIter)
            if (!callSelf)
                call = false;
        if (call) 
            if (!callback(i->item()._node, data))
                if (!searchInRest)
                    break; 
    }
}

void
Node::doWithSiblings(DoWithNodeCallback callback, void *data,
                     bool searchInRest, bool callSelf)
{
    if (hasParent()) {
        Node *parent = getParent();
        int parentField = getParentField();
        FieldValue *parentValue = parent->getField(parentField);
        if (parentValue->getType() == MFNODE) {
            MFNode *value = (MFNode *)parentValue;
            for (int i = 0; value->getSize(); i++) {
                Node *sibling = value->getValue(i);
                if (sibling != NULL) {
                    bool call = true;
                    if (sibling == this)
                        if (!callSelf)
                            call = false;
                    if (call)    
                        if (!callback(sibling, data))
                            if (!searchInRest)
                                break; 
                }
            }      
        }
    }
}

bool Node::doWithBranch(DoWithNodeCallback callback, void *data, 
                        bool searchInRest, bool callSelf)
{
    bool searchOn = true;
    // search in current node itself
    if (callSelf)
        searchOn = callback(this, data);
    if (searchOn) {
        // search in children of of current node
        for (int i = 0; i < _proto->getNumFields(); i++)
            if (getField(i)->getType() == SFNODE) {
                Node *child = ((SFNode *) getField(i))->getValue();
                if (child)
                    if (!child->doWithBranch(callback, data, false))
                        return false;
            } else if (getField(i)->getType() == MFNODE) {
                NodeList *childList = ((MFNode *) getField(i))->getValues();
                if (childList) {
                    for (int j = 0; j < childList->size(); j++) {
                        Node *child = childList->get(j);
                        if (child) 
                            if (!child->doWithBranch(callback, data, false))
                                return false;       
                    }
                }
            }
        if (searchInRest) {
            // search in next children of current parent
            if (this != _scene->getRoot())
                if (hasParent()) {
                    Node *parent = getParent();
                    FieldValue *selfField = parent->getField(getParentField()); 
                    if (selfField->getType() == MFNODE) {
                        int index = parent->findChild(this, getParentField());
                        MFNode *mfNode = (MFNode *)selfField;
                        NodeList *childList = mfNode->getValues();
                        if (childList)
                            for (int j = index + 1; j < childList->size(); j++){
                                Node *child = childList->get(j);
                                if (child) 
                                    if (!child->doWithBranch(callback, data, 
                                                             false))
                                        return false;
                            }
                    }
                }
        }
    }
    return searchOn;
}

bool
Node::doWithSimilarBranch(DoWithSimilarBranchCallback callback, 
                          Node *similarNode, void *data)
{
    if (!callback(this, similarNode, data))
        return false;
    if (_numFields != similarNode->getProto()->getNumFields())
        return false;
    for (int i = 0; i < _numFields; i++) {
        if (_fields[i]->getType() == SFNODE) {
            if (similarNode->getField(i)->getType() != SFNODE)
                return false;
            Node *child = ((SFNode *) _fields[i])->getValue();
            Node *node = ((SFNode *) similarNode->getField(i))->getValue(); 
            if (child && node) { 
                if (!child->doWithSimilarBranch(callback, node, data))
                    return false;
            } else if (child || node)
                return false;
        } else if (_fields[i]->getType() == MFNODE) {
            if (similarNode->getField(i)->getType() != MFNODE)
                return false;
            NodeList *childList = ((MFNode *) _fields[i])->getValues();
            NodeList *nodeList = ((MFNode *) 
                                  similarNode->getField(i))->getValues(); 
            if (childList && nodeList) {
                if (childList->size() != nodeList->size())
                    return false;
                for (int j = 0; j < childList->size(); j++) {
                    Node *child = childList->get(j);
                    Node *node = nodeList->get(j);
                    if (child && node) {
                        if (!child->doWithSimilarBranch(callback, node, data)) 
                            return false;
                    } else if (child || node)
                        return false;
                }
            } else if ((childList != NULL) || (nodeList != NULL))
                return false;
        }
    }
    return true;   
}

void
Node::copyChildrenTo(Node *copyedNode)
{
    for (int i = 0; i < _numFields; i++) {
        if (_fields[i]->getType() == SFNODE) {
            Node *child = ((SFNode *) _fields[i])->getValue();
            if (child) {
                Node *copyChild = child->copy();
                copyChild->ref();
                copyedNode->setField(i, new SFNode(copyChild));
                child->copyChildrenTo(copyChild);
            }
        } else if (_fields[i]->getType() == MFNODE) {
            NodeList *childList = ((MFNode *) _fields[i])->getValues();
            if (childList) {
                MFNode *field = new MFNode((MFNode *)_fields[i]);
                if (field) {
                    copyedNode->setField(i, field);
                    MFNode *mfnode = (MFNode *) copyedNode->getField(i);
                    for (int j = 0; j < childList->size(); j++) {
                        Node *copyChild = mfnode->getValue(j);
                        if (copyChild && childList->get(j)) {
                            childList->get(j)->ref();
                            childList->get(j)->copyChildrenTo(copyChild);
                        }
                    }
                }
            }
        }
    }   
}

void
Node::doWithAllElements(DoWithAllElementsCallback callback, void *data)
{
    int i = 0;
    for (i = 0; i < _proto->getNumFields(); i++)
        callback(_proto->getField(i), data);
    for (i = 0; i < _proto->getNumEventIns(); i++)
        callback(_proto->getEventIn(i), data);
    for (i = 0; i < _proto->getNumEventOuts(); i++)
        callback(_proto->getEventOut(i), data);
    for (i = 0; i < _proto->getNumExposedFields(); i++)
        callback(_proto->getExposedField(i), data);
}

void
NodeData::removeChildren(void) 
{
    for (int i = 0; i < _numFields; i++)
        if (_fields[i]->getType() == SFNODE)
            setField(i, new SFNode(NULL));
        else if (_fields[i]->getType() == MFNODE)
            setField(i, new MFNode());
}

void
NodeData::handleIs(void)
{
    // PROTO initialisation ready ?
    if (_nodePROTO == NULL)
        return;
    // handle IS
    int i;
    for (i = 0; i < _proto->getNumExposedFields(); i++) {
        ExposedField *field = _proto->getExposedField(i);
        if (field->getFlags() & FF_IS)
            for (int j = 0; j < field->getNumIs(); j++) {
                Node *isNode = getIsNode(field->getIsNodeIndex(j));
                FieldValue *value = field->getValue()->copy();
                isNode->setField(field->getIsField(j), value);
            }
    }
    for (i = 0; i < _proto->getNumFields(); i++) {
        Field *field = _proto->getField(i);
        if (field->getFlags() & FF_IS)
            for (int j = 0; j < field->getNumIs(); j++) {
                Node *isNode = getIsNode(field->getIsNodeIndex(j));
                FieldValue *value = field->getDefault()->copy();
                isNode->setField(field->getIsField(j), value);
            }                                 
    }
    if (_proto->isPROTO())
        for (i = 0; i < _proto->getNumEventOuts(); i++) {
            EventOut *eventOut = _proto->getEventOut(i);
            if (eventOut->getFlags() & FF_IS)
               for (int j = 0; j < eventOut->getNumIs(); j++)
                   if (eventOut->getFlags() & EOF_IS) {
                       Node *isNode = getIsNode(eventOut->getIsNodeIndex(j)); 
                       Proto *proto = isNode->getProto();
                       int evOut = eventOut->getIsField(j);
                       EventOut *tEventOut = proto->getEventOut(evOut);
                       tEventOut->addIs((Node *)this, i, EL_EVENT_OUT, NULL, -1,
                                        EOF_IS_HIDDEN);
                   }
        }
}

Node *
NodeData::getIsNode(int nodeIndex)
{
    NodePROTO *nodeProto = (NodePROTO *)this;
    Node *nodeInPROTO =  nodeProto->getIsNode(nodeIndex);
    if (nodeInPROTO)
        return nodeInPROTO;
    assert(0);
    return NULL;
}

bool                
NodeData::isDefault(int field) const
{
    return _fields[field]->equals(getProto()->getField(field)->getDefault());
}

bool 
NodeData::hasInputs(void)
{
    for (int j = 0; j < _proto->getNumEventIns(); j++)
        if (getInput(j).size() > 0)
            return true;
    return false;
}

bool 
NodeData::hasOutputs(void)
{
    for (int j = 0; j < _proto->getNumEventOuts(); j++)
        if (getOutput(j).size() > 0)
            return true;
    return false;
}

bool 
NodeData::hasInput(const char* routename) const
{
    bool x3d = _scene->isX3d();
    for (int i = 0; i < _proto->getNumEventIns(); i++)
        if (strcmp(_proto->getEventIn(i)->getName(x3d), routename) == 0)
            if (getInput(i).size() > 0)
                return true;
            else
                return false;
    assert(0);
    return false;
}

bool 
NodeData::hasOutput(const char* routename) const
{
    bool x3d = _scene->isX3d();
    for (int i = 0; i < _proto->getNumEventOuts(); i++)
        if (strcmp(_proto->getEventOut(i)->getName(x3d), routename) == 0)
            if (getOutput(i).size() > 0)
                return true;
            else
                return false;
    assert(0);
    return false;
}

int
NodeData::writeProtoArguments(int f) const
{
    bool x3d = _scene->isX3dv();
    int i;
    for (i = 0; i < _proto->getNumFields(); i++) {
        RET_ONERROR( _proto->getField(i)->writeWithoutValue(f, 1, x3d) )
        TheApp->incSelectionLinenumber();
    }
    for (i = 0; i < _proto->getNumExposedFields(); i++) {
        RET_ONERROR( _proto->getExposedField(i)->writeWithoutValue(f, 1, x3d) )
        TheApp->incSelectionLinenumber();
    }
    for (i = 0; i < _proto->getNumEventIns(); i++) {
        RET_ONERROR( _proto->getEventIn(i)->write(f, 1, x3d) )
        TheApp->incSelectionLinenumber();
    }
    for (i = 0; i < _proto->getNumEventOuts(); i++) {
        RET_ONERROR( _proto->getEventOut(i)->write(f, 1, x3d) )
        TheApp->incSelectionLinenumber();
    }
    return 0;
}

bool
NodeData::hasDefault(int flag)
{
    for (int i = 0; i < _numFields; i++) {
        Field *field = _proto->getField(i);
        if (field->getFlags() & flag)
            if (!(_fields[i]->equals(field->getDefault())))
                return false;
    }
    return true;
}

void
NodeData::generateTreeLabel(void)
{
    _treeLabel = "";
    if (TheApp->is4Catt())
        if (isMesh()) {
            if (hasTwoSides()) {
               if (isDoubleSided())
                   _treeLabel += "(2 sides) ";
               else
                   _treeLabel += "(1 side) ";
            } else
               _treeLabel += "(only 1 side) ";
        }     
    if (hasName()) {
        _treeLabel += getName();
        const char *protoName = _proto->getName(_scene->isX3d());
        if (stringncmp(getName(), protoName) != 0) { 
            _treeLabel += " ";
            _treeLabel += protoName;
        }
    } else
        _treeLabel += _proto->getName(_scene->isX3d());
}

int
Node::getParentIndex(void) const
{
    FieldValue *value = getParentFieldValue();
    if (value == NULL)
        return -1;
    if (value->getType() == MFNODE) {
        MFNode *mfValue = (MFNode *)value;
        for (int i = 0; i < mfValue->getSize(); i++) {
            Node *node = mfValue->getValue(i);
            if (node->isEqual((Node *)this))
                if (_geometricParentIter == node->getGeometricParentIter())
                    return i;
            }
    }
    return -1;
}

Node *
Node::getParent(int nodeType) const
{
    for (ParentIter i = _parents.first(); i != NULL; i = i->next()) {
        Node *node = i->item()._node;
        if (node->getType() == nodeType)
            return node;
    }
    return getParent();
}

FieldValue *
Node::getParentFieldValue(void) const
{
    if (!hasParent())
        return NULL;
    Node *parent = getParent();
    int parentField = getParentField();
    return parent->getField(parentField);
}

int
Node::getPrevSiblingIndex(void)
{
    int parentIndex = getParentIndex();
    if (parentIndex > 0)
        return parentIndex - 1;
    return -1;
}

int
Node::getNextSiblingIndex(void)
{
    FieldValue *value = getParentFieldValue();
    if (value == NULL)
        return -1;
    if (value->getType() == MFNODE) {
        MFNode *mfValue = (MFNode *)value;
        for (int i = 0; i < mfValue->getSize() - 1; i++) {
            Node *node = mfValue->getValue(i);
            if (node->isEqual((Node *)this))
                if (_geometricParentIter == node->getGeometricParentIter())
                    return i + 1;
            }
    }
    return -1;
}

Node *
NodeData::convert2X3d(void) {
    for (int i = 0; i < _numFields; i++)
        if (_fields[i] != NULL)
           if (_fields[i]->getType() == SFNODE)
               ((SFNode *) _fields[i])->convert2X3d();
           else if (_fields[i]->getType() == MFNODE)
               ((MFNode *) _fields[i])->convert2X3d();
    return NULL;
}

Node *
NodeData::convert2Vrml(void) {
    for (int i = 0; i < _numFields; i++)
        if (_fields[i] != NULL)
           if (_fields[i]->getType() == SFNODE)
               ((SFNode *) _fields[i])->convert2Vrml();
           else if (_fields[i]->getType() == MFNODE)
               ((MFNode *) _fields[i])->convert2Vrml();
    return NULL;
}

bool
NodeData::canWriteAc3d(void)
{ 
    for (int i = 0; i < _numFields; i++)
        if (_fields[i] != NULL)
           if (_fields[i]->getType() == SFNODE) {
               Node *node = ((SFNode *)_fields[i])->getValue();
               if (node)
                   return node->canWriteAc3d();
           } else if (_fields[i]->getType() == MFNODE) {
               MFNode *nodes = (MFNode *) _fields[i];
               for (int j = 0; j < nodes->getSize(); j++)
                   if (nodes->getValue(j))
                       if (nodes->getValue(j)->canWriteAc3d())
                           return true; 
           }     
    return false; 
}

int
NodeData::writeAc3d(int f, int indent)
{ 
    for (int i = 0; i < _numFields; i++)
        if (_fields[i] != NULL)
           if (_fields[i]->getType() == SFNODE)
               RET_ONERROR( ((SFNode *) _fields[i])->writeAc3d(f, indent) )
           else if (_fields[i]->getType() == MFNODE)
               RET_ONERROR( ((MFNode *) _fields[i])->writeAc3d(f, indent) )
    return 0; 
}

int
NodeData::writeAc3dMaterial(int &materialIndex, int f, int indent) 
{ 
    for (int i = 0; i < _numFields; i++)
        if (_fields[i] != NULL)
           if (_fields[i]->getType() == SFNODE) {
               SFNode *field = (SFNode *) _fields[i];
               RET_ONERROR( field->writeAc3dMaterial(materialIndex, f, indent) )
           } else if (_fields[i]->getType() == MFNODE) {
               MFNode *field = (MFNode *) _fields[i];
               RET_ONERROR( field->writeAc3dMaterial(materialIndex, f, indent) )
           }
    return 0;
}

bool
Node::canWriteCattGeo(void)
{ 
    for (int i = 0; i < _numFields; i++)
        if (_fields[i] != NULL)
           if (_fields[i]->getType() == SFNODE) {
               Node *node = ((SFNode *)_fields[i])->getValue();
               if (node)
                   return node->canWriteCattGeo();
           } else if (_fields[i]->getType() == MFNODE) {
               MFNode *nodes = (MFNode *) _fields[i];
               for (int j = 0; j < nodes->getSize(); j++)
                   if (nodes->getValue(j))
                       if (nodes->getValue(j)->canWriteCattGeo())
                           return true; 
           }     
    return false; 
}

int
Node::writeCattGeo(int f, int indent)
{ 
    for (int i = 0; i < _numFields; i++)
        if (_fields[i] != NULL)
           if (_fields[i]->getType() == SFNODE)
               RET_ONERROR( ((SFNode *) _fields[i])->writeCattGeo(f, indent) )
           else if (_fields[i]->getType() == MFNODE)
               RET_ONERROR( ((MFNode *) _fields[i])->writeCattGeo((Node *)this, 
                                                                  f, indent) )
    return 0; 
}

