/*
 * NodeMaterial.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include "stdafx.h"

#include "NodeMaterial.h"
#include "Proto.h"
#include "FieldValue.h"
#include "ExposedField.h"
#include "SFFloat.h"
#include "SFColor.h"
#include "Field.h"
#include "ExposedField.h"
#include "Util.h"
#include "Scene.h"

ProtoMaterial::ProtoMaterial(Scene *scene)
  : Proto(scene, "Material")
{
    ambientIntensity.set(
          addExposedField(SFFLOAT, "ambientIntensity", new SFFloat(0.2f), 
                          new SFFloat(0.0f), new SFFloat(1.0f)));
    diffuseColor.set(
          addExposedField(SFCOLOR, "diffuseColor", 
                          new SFColor(0.8f, 0.8f, 0.8f)));
    setFieldFlags(diffuseColor, EIF_RECOMMENDED);
    emissiveColor.set(
          addExposedField(SFCOLOR, "emissiveColor", 
                          new SFColor(0.0f, 0.0f, 0.0f)));
    shininess.set(
          addExposedField(SFFLOAT, "shininess", new SFFloat(0.2f),
                          new SFFloat(0.0f), new SFFloat(1.0f)));
    specularColor.set(

          addExposedField(SFCOLOR, "specularColor", 
                          new SFColor(0.0f, 0.0f, 0.0f)));
    transparency.set(
          addExposedField(SFFLOAT, "transparency", new SFFloat(0.0f), 
                          new SFFloat(0.0f), new SFFloat(1.0f)));
    setFieldFlags(transparency, EIF_RECOMMENDED);
/*
    mirror.set(
          addExposedField(SFFLOAT, "mirror", new SFFloat(0.0f), 
                          new SFFloat(0.0f), new SFFloat(1.0f)));
    setFieldFlags(mirror, FF_KAMBI_ONLY);
*/
}

Node *
ProtoMaterial::create(Scene *scene)
{ 
    return new NodeMaterial(scene, this);
}

NodeMaterial::NodeMaterial(Scene *scene, Proto *def)
  : Node(scene, def)
{
    _ac3dMaterialIndex = -1;
}


void NodeMaterial::bind()
{
    float ambientColor[4], dc[4], ec[4];

    for (int i = 0; i < 3; i++) {
        ec[i] = emissiveColor()->getValue()[i];
        dc[i] = diffuseColor()->getValue()[i];
        ambientColor[i] = diffuseColor()->getValue()[i] * 
                          ambientIntensity()->getValue();
    }
    ambientColor[3] = 1.0f;
    ec[3] = dc[3] = 1.0f - transparency()->getValue();

    Util::myGlMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambientColor);
    Util::myGlMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, dc);
    Util::myGlMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, ec);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128.0f * 
                                                 shininess()->getValue());
    Util::myGlMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specularColor()->getValue());

    if (transparency()->getValue() > 0.0f) {
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glEnable(GL_BLEND);
    }
}

void NodeMaterial::unbind()
{
    float ambientColor[4];

    for (int i = 0; i < 3; i++) {
        ambientColor[i] = diffuseColor()->getValue()[i] * 
                          ambientIntensity()->getValue();
    }
    ambientColor[3] = 1.0f;

    Util::myGlMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambientColor);
    Util::myGlMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuseColor()->getValue());
    Util::myGlMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, emissiveColor()->getValue());
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shininess()->getValue());
    Util::myGlMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specularColor()->getValue());

    glDisable(GL_BLEND);
}

void
NodeMaterial::emissive2diffuse(void)
{
    SFColor *color = new SFColor(emissiveColor()->getValue(0),
                                 emissiveColor()->getValue(1),
                                 emissiveColor()->getValue(2));
    _scene->backupField(this, diffuseColor_Field()); 
    _scene->setField(this, diffuseColor_Field(), color);
    color = new SFColor(0, 0, 0);
    _scene->backupField(this, emissiveColor_Field()); 
    _scene->setField(this, emissiveColor_Field(), color);
}

void
NodeMaterial::diffuse2emissive(void)
{
    SFColor *color = new SFColor(diffuseColor()->getValue(0),
                                 diffuseColor()->getValue(1),
                                 diffuseColor()->getValue(2));
    _scene->backupField(this, emissiveColor_Field()); 
    _scene->setField(this, emissiveColor_Field(), color);
    color = new SFColor(0, 0, 0);
    _scene->backupField(this, diffuseColor_Field()); 
    _scene->setField(this, diffuseColor_Field(), color);
}

bool    
NodeMaterial::isTransparent(void)
{
    if (transparency()->getValue() > 0.0)
        return true;
    else
        return false;
}

int 
NodeMaterial::getProfile(void) const
{ 
    if (!isDefault(ambientIntensity_Field()))
        return PROFILE_IMMERSIVE;
    if (!isDefault(shininess_Field()))
        return PROFILE_IMMERSIVE;
    if (!isDefault(specularColor_Field()))
        return PROFILE_IMMERSIVE;
    if (!isDefault(emissiveColor_Field()))
        return PROFILE_IMMERSIVE;
    return PROFILE_INTERCHANGE;
}

int 
NodeMaterial::writeAc3dMaterial(int &materialIndex, int f, int indent)
{
    _ac3dMaterialIndex = materialIndex++;
    RET_ONERROR( mywritef(f, "MATERIAL \"") )
    RET_ONERROR( writeExportMaterialName(f) )
    RET_ONERROR( mywritef(f, "\" ") ) 
    RET_ONERROR( mywritef(f, "rgb %f %f %f  ", 
                          diffuseColor()->getValue(0),
                          diffuseColor()->getValue(1),
                          diffuseColor()->getValue(2)) )
    RET_ONERROR( mywritef(f, "amb %f %f %f  ",
                          ambientIntensity()->getValue(),
                          ambientIntensity()->getValue(),
                          ambientIntensity()->getValue()) )
    RET_ONERROR( mywritef(f, "emis %f %f %f  ", 
                          emissiveColor()->getValue(0),
                          emissiveColor()->getValue(1),
                          emissiveColor()->getValue(2)) )
    RET_ONERROR( mywritef(f, "spec %f %f %f  ", 
                          specularColor()->getValue(0),
                          specularColor()->getValue(1),
                          specularColor()->getValue(2)) )
    RET_ONERROR( mywritef(f, "shi %d  ", ((int)(shininess()->getValue() * 128))))
    RET_ONERROR( mywritef(f, "trans %f\n", transparency()->getValue()) )
    return 0;
}

/*
    In VRML97/X3D the name of a material is simply the DEF name of the 
    Material node, so the following function is not needed when dealing
    with VRML data.
    Instead, this function is used when exporting to non VRML/X3D files.
    Some export file formats are used for sound simulation, where 
    the material name is used for applying sound related parameters to
    geometry. 
    In sound simulation it is important to have a easy way to apply
    sound parameters to all surfaces. Therefore the following function
    search for other DEF names in higher levels of the scenegraph, if
    the current node do not have a DEF name.
    Cause some 3D tools write hard to use DEF names, a preference can be
    used to select, if the part before the first _ or the part after the
    last _ or the full DEF name is used for material name generation
 */

int
NodeMaterial::writeExportMaterialName(int f, const char *defaultName)
{
    Node *node = this;   
    // search for DEF name till root of scenegraph
    while (1) {
        if (node->hasName()) {
            const char *name = node->getName();
            if (TheApp->IsMaterialNameBeforeUnderscore()) {
                char *firstPartOfName = strdup(name);
                char *endPointer = (char *)strchr(firstPartOfName, '_');
                if (endPointer) {
                    endPointer[0] = 0;
                    if (strlen(firstPartOfName) > 0) {
                        RET_ONERROR( mywritef(f, "%s", firstPartOfName) )
                        return 0;
                    } else if (defaultName != NULL) {
                        RET_ONERROR( mywritef(f, "%s", defaultName) )
                        return 0;
                    }
                } else
                    RET_ONERROR( mywritef(f, "%s", name) )
                free(firstPartOfName);
            } else if (TheApp->IsMaterialNameAfterUnderscore()) {
                const char *lastPartOfName = strrchr(name, '_');
                if (lastPartOfName) {
                    if (strlen(lastPartOfName + 1) > 0) {
                        RET_ONERROR( mywritef(f, "%s", lastPartOfName + 1) )
                        return 0;
                    } else if (defaultName != NULL) {
                        RET_ONERROR( mywritef(f, "%s", defaultName) )
                        return 0;
                    }
                } else {
                    RET_ONERROR( mywritef(f, "%s", name) )
                    return 0;
                }
            } else
                RET_ONERROR( mywritef(f, "%s", name) )      
            return 0;
        } else {
            if (node->hasParent())
                node = node->getParent();
            else {
               if (defaultName != NULL)
                   RET_ONERROR( mywritef(f, "%s", defaultName) )
               break;
            }           
        }
    }
    return 0;
}


