/*
 * Proto.h
 *
 * Copyright (C) 1999 Stephen F. White, 2005 J. "MUFTI" Scheurich
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#ifndef _PROTO_H
#define _PROTO_H

#ifndef _ARRAY_H
# include "Array.h"
#endif
#ifndef _MAP_H
# include "Map.h"
#endif
#ifndef _DUNE_STRING_H
# include "MyString.h"
#endif
#ifndef _NODE_H
# include "Node.h"
#endif

#define INVALID_INDEX (-1)

class Element;
class EventIn;
class EventOut;
class Field;
class ExposedField;
class FieldValue;
class Scene;
class Node;
#include "StringArray.h"
#include "NodeList.h"

class NodeScript;
class ScriptDialog;

class Proto {
public:

    // unlike other nodes, script nodes need to change their Proto dynamically
    friend class NodeScript;
    friend class ScriptDialog;
    friend class DynamicFieldsNode;

                          Proto(Scene *scene, const MyString &name);
                          // create a Proto as a X3D/Cover extension
                          Proto(Scene *scene, Proto *proto, int extensionFlag);
    virtual              ~Proto();

    FieldIndex metadata;

    void                  protoInitializer(Scene *scene, const MyString &name);

    bool                  avoidElement(Element *element, int flag);
    virtual Node         *create(Scene *scene);

    EventIn              *getEventIn(int index) const 
                             { return _eventIns[index]; }
    EventOut             *getEventOut(int index) const 
                             { return _eventOuts[index]; }
    Field                *getField(int index) const { return _fields[index]; }
    ExposedField         *getExposedField(int index) const 
                             { return _exposedFields[index]; }

    int                   getNumFields() const { return _fields.size(); }
    int                   getNumEventIns() const { return _eventIns.size(); }
    int                   getNumEventOuts() const { return _eventOuts.size(); }
    int                   getNumExposedFields() const 
                             { return _exposedFields.size(); }

    int                   lookupEventIn(const MyString &name, 
                                        bool x3d = false) const;
    int                   lookupEventOut(const MyString &name, 
                                         bool x3d = false) const;
    int                   lookupField(const MyString &name, 
                                      bool x3d = false) const;
    int                   lookupExposedField(const MyString &name,
                                             bool x3d = false) const;
    int                   lookupIsEventIn(Node *node, int eventIn,
                                          int elementType = -1) const;
    int                   lookupIsEventOut(Node *node, int eventOut,
                                           int elementType = -1) const;
    int                   lookupIsField(Node *node, int field) const;
    int                   lookupIsExposedField(Node *node, int field) const;

    virtual const MyString &getName(bool x3d) const { return _name; }

    bool                  canWriteElement(Element *element, bool x3d) const;
    int                   write(int filedes, int indent, bool x3d) const;
    int                   writeEvents(int filedes, int indent, bool x3d, 
                                      bool withFields = true,
                                      bool withValues = true) const;

    int                   addElement(Element *element);
    int                   addField(Field *field);
    int                   addExposedField(ExposedField *exposedField);

    void                  define(Node *primaryNode, NodeList *nodes);
    bool                  isPROTO(void);
    int                   getNumNodes(void) { return _protoNodes.size(); }
    Node                 *getNode(int i) { return _protoNodes[i]; }
    void                  compareToIsNodes(Node *node);
    void                  addURLs(FieldValue *urls);

    bool                  isX3dProto(void) 
                             { return isExtensionProto(FF_X3D_ONLY); }
    bool                  isCoverProto(void) 
                             { return isExtensionProto(FF_COVER_ONLY); }
    bool                  isKambiProto(void) 
                             { return isExtensionProto(FF_KAMBI_ONLY); }
    bool                  isExtensionProto(int flag);
    bool                  checkIsExtension(Element *element, int flag);
    void                  setScene(Scene *scene) { _scene = scene; }

    void                  convert2X3d(void) 
                             {
                             for (int i = 0; i < _protoNodes.size(); i++)
                                _protoNodes[i]->convert2X3d();
                             }

    void                  convert2Vrml(void) 
                             {
                             for (int i = 0; i < _protoNodes.size(); i++)
                                _protoNodes[i]->convert2X3d();
                             }

protected:
    Proto                *copy() const { return new Proto(*this); }
    int                   addField(int fieldType, const MyString &name,
                                   FieldValue *defaultValue,
                                   FieldValue *min = NULL,
                                   FieldValue *max = NULL);
    int                   addField(int fieldType, const MyString &name,
                                   FieldValue *defaultValue, int nodeType);
    int                   addField(int fieldType, const MyString &name,
                                   FieldValue *defaultValue, int flags,
                                   const char **strings);
    int                   addEventIn(int fieldType, const MyString &name);
    int                   addEventIn(int fieldType, const MyString &name,
                                     int flags);
    int                   addEventIn(int fieldType, const MyString &name,
                                     int flags, int field);
    int                   addEventOut(int fieldType, const MyString &name);
    int                   addEventOut(int fieldType, const MyString &name,
                                      int flags);
    int                   addExposedField(int fieldType, const MyString &name,
                                          FieldValue *defaultValue,
                                          FieldValue *min = NULL,
                                          FieldValue *max = NULL,
                                          const MyString &x3dName = "");
    int                   addExposedField(int fieldType, const MyString &name,
                                          FieldValue *defaultValue, 
                                          int nodeType,
                                          const MyString &x3dName = "");
    int                   addExposedField(int fieldType, const MyString &name,
                                          FieldValue *defaultValue, int flags,
                                          const char **strings);
    void                  deleteEventOut(int i);
    int                   lookupSimpleEventIn(const MyString &name, 
                                              bool x3d) const;
    int                   lookupSimpleEventOut(const MyString &name,
                                               bool x3d) const;
    void                  setIsNodeIndex(void);
    void                  setFieldFlags(FieldIndex index, int flags);

    void                  handleVrmlCut(EventOut *eventOut);

protected:
    Scene                *_scene;
    MyString              _name;
    Array<EventIn *>      _eventIns;
    Array<EventOut *>     _eventOuts;
    Array<Field *>        _fields;
    Array<ExposedField *> _exposedFields;

    // for PROTO's:
    Array<Node *>         _protoNodes;
    int                   _nodeIndex;
    FieldValue           *_urls;
};

class NodePROTO : public Node
{
public:
                        NodePROTO(Scene *scene, Proto *proto);

    virtual int         getType() const; 
    virtual int         getNodeClass() const;
    virtual int         getProfile(void) const;
    virtual Node       *copy() const { return new NodePROTO(*this); }

    virtual Proto      *getPrimaryProto() const
                           { return _indexedNodes[0]->getProto(); }

    virtual void        preDraw();
    virtual void        draw();
    virtual void        draw(int pass);

    virtual FieldValue *getField(int index) const;
    virtual FieldValue *getField(FieldIndex index) const;
    virtual void        setField(int index, FieldValue *value);
    virtual void        setField(FieldIndex index, FieldValue *value);
    virtual void        sendEvent(int eventOut, double timestamp, 
                                  FieldValue *value);
    virtual void        receiveEvent(int eventIn, double timestamp, 
                                     FieldValue *value);
    void                buildIndexedNodes(void);
    Node               *getIsNode(int nodeIndex);
    void                appendToIndexedNodes(Node *node);
    virtual bool        isJoint(void) { return _jointRotationField != -1; }
    virtual bool        isHumanoid(void) { return _isHumanoid; }
    virtual void        drawHandles();
    virtual int         countPolygons(void);
    virtual int         countPrimitives(void);
    virtual int         countPolygons1Sided(void);
    virtual int         countPolygons2Sided(void);
    virtual Vec3f       getHandle(int handle, int *constraint, int *field);
    virtual void        setHandle(int handle, const Vec3f &v);
    virtual void        transform();
    virtual bool        isPROTO() { return true; }
    void                getComponentsInBranch(DoWithNodeCallback callback,
                                              void *data);
    virtual bool        canWriteAc3d();
    virtual int         writeAc3d(int filedes, int indent);
    virtual int         writeAc3dMaterial(int &materialIndex, int filedes, 
                                          int indent);
    virtual bool        canWriteCattGeo();
    virtual int         writeCattGeo(int filedes, int indent);


protected:
    NodeList            _nodes;
    Array<Node *>       _indexedNodes;
    int                 _jointRotationField;
    bool                _isHumanoid;
};

typedef Array<Proto *>  ProtoArray;

#endif // _PROTO_H
